(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Omega.

Require Import rel_utils.
Require Import list_utils.
Require Import good_base.
Require Import af.
Require Import bar.
Require Import good_bar.

Set Implicit Arguments.

Section af_bar.

  Variable (X : Type).

  Implicit Type (R : X -> X -> Prop).

  Section bar_good_af.

    Let bar_good_af_rec R ll : bar (good R) ll -> af (lift_rel_list R ll).
    Proof.
      induction 1 as [ ll Hll | ll Hll IH ].
      
      apply in_af_0.
      induction Hll as [ ll a b Hll Hab | ll a Hll IH ] .
      intros; simpl; right; apply lift_rel_list_In with (1 := Hab); auto.
      intros; simpl; left; auto.
      
      apply in_af_1, IH.
    Qed.

    Let af_bar_good_rec R : af R -> forall S ll, R inc2 lift_rel_list S ll -> bar (good S) ll.
    Proof.
      induction 1 as [ R HR | R HR IHR ]; intros S ll H.
      constructor 2; intros u.
      constructor 2; intros v.
      constructor 1.
      cutrewrite (v::u::ll = (v::u::nil)++ll); auto.
      apply good_lift_rel_list.
      apply good_inc with (1 := H).
      constructor 1 with u; auto; left; auto.
      constructor 2; intros a.
      apply (IHR a _ (a::ll)).
      simpl; unfold lift_rel.
      intros ? ? [ | ]; [ left | right ]; auto.
    Qed.

    Theorem af_bar_eq R ll : af (lift_rel_list R ll) <-> bar (good R) ll. 
    Proof.
      split; auto.
      intros H.
      apply af_bar_good_rec with (1 := H); auto.
    Qed.

  End bar_good_af.

  Theorem af_eq_bar_good R : af R <-> bar (good R) nil.
  Proof.
    apply af_bar_eq with (ll := nil).
  Qed.

End af_bar.

Fact bar_ramsey X Y R S : bar (@good X R) nil 
                       -> bar (@good Y S) nil 
                       -> bar (good (rel_prod R S)) nil.
Proof.
  do 3 rewrite <- af_eq_bar_good.
  intros H1 H2.
  generalize (af_product H1 H2).
  apply af_inc.
  intros [] []; auto.
Qed.

