(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import sublist.
Require Import subword.

Require Import continuity.

Require Import tree.

Set Implicit Arguments.

Section emb_tree_homeo.

  Variables (X : Type) (R : X -> X -> Prop). 
  
  Reserved Notation "x '<eh' y" (at level 70, no associativity).

  (*

  Inductive emb_tree_homeo : tree X -> tree X -> Prop :=
      | in_embth_0 : forall s t x ll, In t ll -> s <eh t -> s <eh in_tree x ll
      | in_embth_1 : forall x y ll mm kk, R x y -> subword emb_tree_homeo mm kk -> in_tree x ll <eh in_tree y kk
    where "x <eh y" := (emb_tree_homeo x y).
    
  *)
 
  Let embth_fop (S : relation (tree X)) t1 t2 := 
      (exists t a ll, In t ll /\ S t1 t /\ t2 = in_tree a ll )
   \/ (exists a b ll mm, R a b /\ subword S ll mm /\ t1 = in_tree a ll /\ t2 = in_tree b mm).

  Let embth_fop_monotonic : op_monotonic embth_fop.
  Proof.
    intros S T H x y [ (t & a & ll & H1 & H2 & ?)
                     | (a & b & ll & mm & H1 & H2 & ? & ?) ]; subst.
    left; exists t, a, ll; auto.
    right; exists a, b, ll, mm; repeat split; auto.
    revert H2; apply subword_inc; auto.
  Qed.

  Let embth_fop_w_continuous : op_w_continuous embth_fop.
  Proof.
    intros K HK x y  [ (t & a & ll & H1 & H2 & ?)
                     | (a & b & ll & mm & H1 & H2 & ? & ?) ]; subst.

    destruct H2 as (n & H2); exists n; left; exists t, a, ll; auto.

    apply sw_w_continuous in H2; auto. 
    destruct H2 as (n & Hn).
    exists n; right; exists a, b, ll, mm; auto.
  Qed.

  Definition emb_tree_homeo := lfp embth_fop.

  Infix "<eh" := emb_tree_homeo (at level 70).

  Let embth_fix t1 t2 : t1 <eh t2 <-> (exists t a ll, In t ll /\ t1 <eh t /\ t2 = in_tree a ll )
                                   \/ (exists a b ll mm, R a b /\ subword emb_tree_homeo ll mm /\ t1 = in_tree a ll /\ t2 = in_tree b mm).
  Proof.
    apply lfp_fixpoint_eq; auto.
  Qed.

  Fact emb_tree_homeo_inv x ll y mm : in_tree x ll <eh in_tree y mm <-> (exists t, In t mm /\ in_tree x ll <eh t)
                                                                     \/ (R x y /\ subword emb_tree_homeo ll mm).
  Proof.
    rewrite embth_fix; split.

    intros [ (t & a & ll' & H1 & H2 & H3)
           | (a & b & ll' & mm' & H1 & H2 & H3 & H4) ]; subst.
    injection H3; intros; subst; left; exists t; auto.
    injection H3; injection H4; intros; subst; right; auto.
    
    intros [ (t & H1 & H2) 
           | (H1 & H2) ].
    left; exists t, y, mm; auto.
    right; exists x, y, ll, mm; auto.
  Qed. 
    
  (* Let us show that emb_tree_homeo is indeed the least relation closed under the
     intended inductive rules *)

  Section emb_tree_homeo_closed.
  
    Fact in_emb_tree_homeo_0 t t' a ll : In t' ll -> t <eh t' -> t <eh in_tree a ll.
    Proof.
      rewrite (embth_fix _ (in_tree _ _)).
      left; exists t', a, ll; tauto.
    Qed.

    Fact in_emb_tree_homeo_1 a b ll mm : R a b -> subword emb_tree_homeo ll mm -> in_tree a ll <eh in_tree b mm.
    Proof.
      rewrite (embth_fix _ (in_tree _ _)).
      right; exists a, b, ll, mm; tauto.
    Qed.
  
  End emb_tree_homeo_closed.

  (* Hence we have the 2 constructors and the most general induction principle emb_tree_homeo_ind *)

  Section emb_tree_homeo_ind.

    Variable S : tree X -> tree X -> Prop.

    Infix "<<" := S (at level 70).

    Hypothesis S_sub0 : forall t t' a ll, In t' ll -> t <eh t' -> t << t' -> t << in_tree a ll.
    Hypothesis S_sub1 : forall a b ll mm, R a b -> subword emb_tree_homeo ll mm -> subword S ll mm -> in_tree a ll << in_tree b mm.

    Fact emb_tree_homeo_ind : forall t1 t2, t1 <eh t2 -> S t1 t2.
    Proof.
      apply lfp_ind; auto.     
      intros x y [ (t & a & ll' & H1 & (H2 & H3) & H4)
                 | (a & b & ll' & mm' & H1 & H2 & H3 & H4) ]; subst.
      apply S_sub0 with (1 := H1); auto.
      apply S_sub1 with (1 := H1); auto; revert H2; apply sw_monotonic; tauto.
    Qed.
     
  End emb_tree_homeo_ind.

  Fact emb_homeo_sub r s t : r <eh s -> s <st t -> r <eh t.
  Proof.
    intros Hrs.
    induction 1 as [ | ? ? ? ? H ]; auto.
    apply in_emb_tree_homeo_0 with (1 := H); auto.
  Qed.

  Fact sub_emb_tree_homeo r s t : r <st s -> s <eh t -> r <eh t.
  Proof.
    intros H.
    induction 1 as [ s t x ll H1 H2 IH2 
                   | x y ll mm H1 H2 IH2 ] using emb_tree_homeo_ind.
    apply in_emb_tree_homeo_0 with (1 := H1); auto.
    apply sub_tree_inv in H.
    destruct H as [ H | H ].
    subst.
    apply in_emb_tree_homeo_1; auto.
    
    apply ssub_tree_inv_right in H.
    destruct H as (s & H3 & z & kk & H4 & H5).
    injection H5; clear H5; intros ? ?; subst z kk.
    apply in_split in H4; destruct H4 as (u & v & H4); subst ll.
    apply subword_app_inv in IH2.
    destruct IH2 as (c & d & ? & ? & IH2); subst.
    apply subword_cons_inv in IH2.
    destruct IH2 as (l & e & k & ? & IH2 & ?); subst.
    apply in_emb_tree_homeo_0 with e; auto.
    rewrite <- app_ass.
    apply in_or_app; right; simpl; auto.
  Qed.    

  Section emb_tree_homeo_dec.

    Hypothesis R_dec : forall x y, { R x y } + { ~ R x y }.
    
    Definition emb_tree_homeo_dec t1 t2 : { t1 <eh t2 } + { ~ t1 <eh t2 }.
    Proof.
      revert t1; induction t2 as [ y l2 IH ]; intros t1.
      
      destruct (@list_dec_rect _ (fun x => t1 <eh x) (fun x => ~ t1 <eh x) l2) as [ (t & H1 & H2) | H ].      
      intros x Hx; destruct (IH x Hx t1); tauto.
      
      left; apply in_emb_tree_homeo_0 with t; auto.
      apply In_t_In; auto.
      
      destruct t1 as [ x l1 ].
      destruct (R_dec x y) as [ Hxy | Hxy ].
      
      destruct (subword_dec emb_tree_homeo l1 l2) as [ Hl12 | Hl12 ].
      intros a b Ha Hb; apply IH; auto.
      
      left; apply in_emb_tree_homeo_1; auto.
      
      right.
      contradict Hl12.
      rewrite emb_tree_homeo_inv in Hl12.
      destruct Hl12 as [ (t & H1 & H2) | (H1 & H2) ]; auto.
      exfalso; rewrite In_In_t in H; apply H in H1; tauto.
      
      right.
      contradict Hxy.
      rewrite emb_tree_homeo_inv in Hxy.
      destruct Hxy as [ (t & H1 & H2) | (H1 & H2) ]; auto.
      exfalso; rewrite In_In_t in H; apply H in H1; tauto.
    Qed.

  End emb_tree_homeo_dec.
  
End emb_tree_homeo.
