(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import tree.

Set Implicit Arguments.

Section tree_kruskal_fix.

  Variable X : Type.

  (* This is a T := ptree_fix is a fixpoint which verifies that T contains isomorphic copies of 

     a/ X
     b/ T + T * list (list (tree T))
   
     These three are not needed 

     c/ T * list (list (tree T))
     d/ T + nat * T * tree T 
     e/ nat * T * tree T

     But instead we need a fixed value in T that serve as undefined output for partial
     functions

  *)

  (* we do not need a least fixpoint so induction principles are not needed.
     Anyway, they would not be correctly derived by the following definition
  *)

  Unset Elimination Schemes.
  
  Inductive ktree_fix :=
    | in_ktf_u : ktree_fix
    | in_ktf_0 : X -> ktree_fix
    | in_ktf_1 : ktree_fix -> ktree_fix
 (*   | in_ktf_2 : nat -> ktree_fix -> tree ktree_fix -> ktree_fix   *)
    | in_ktf_3 : ktree_fix -> list (list (tree ktree_fix)) -> ktree_fix.
    
  Set Elimination Schemes.

End tree_kruskal_fix.

Implicit Arguments in_ktf_u [ X ].
    